C/* 
C * This version looks for regular increment patterns in the ja array.
C * The ka array basically gives the diagonal blocks of the matrix.
C * The data structure is otherwise AIJ; in particular, data has NOT
C * been re-ordered to improve diagonal access.
C */
      subroutine mv( ia, ja, ka, a, x, y, n )
      implicit none
      integer i, jrow, j, ncol, n
      integer k, cnt, kk, col, acol, bsize
      integer ia(*), ja(*), ka(*)
      double precision sum, col2, s1, s2
      double precision a(*), x(n), y(n)
C
      i = 1
      cnt = 0
C
C Use nonstandard do while
C
      do while (i.le.n)
         jrow  = ia(i)
         ncol  = ia(i+1) - jrow
         cnt = cnt+1
         k	  = ka(cnt)
         bsize = k - i + 1
         if (bsize .eq. 1) then
C           /* If a single row, drop to regular code */
	   sum  = 0.0
           do j = 1,ncol
	       sum = sum + a(jrow)*x(ja(jrow))
               jrow = jrow + 1
           enddo
	   y(i) = sum
 	   i = i + 1
        else 
C            /* Switch to diagonal code */
           do kk = 0, bsize-1
	       y(i+kk) = 0.0
           enddo
           do j = 0,ncol-2,2
	       acol = jrow+j
	       col  = ja(acol)
	       col2 = ja(acol+1)
               do kk = 0,bsize-1,2
                 s1 = y(i+kk)
                 s2 = y(i+kk+1)
                 s1 = s1 + a(acol)*x(col+kk) + a(acol+1)*x(col2+kk)
		 acol    = acol +  ncol
                 s2 = s2 + a(acol)*x(col+kk+1) + a(acol+1)*x(col2+kk+1)
		 acol    = acol +  ncol
	         y(i+kk) = s1
	         y(i+kk+1) = s2
               enddo
               if  (kk.lt.bsize) then
                 y(i+kk) = y(i+kk) + a(acol)*x(col+kk)
               endif
	   enddo
           if (j .lt. ncol) then
            acol = jrow+j
            col  = ja(acol)
            do kk=0,bsize-2,2 
                s1 = y(i+kk) 
                s2 = y(i+kk+1)
                s1 = s1 + a(acol) * x(col+kk)
                acol  = acol + ncol
                s2 = s2 + a(acol) * x(col+kk+1)
                acol  = acol + ncol
                y(i+kk) = s1
                y(i+kk+1) = s2
            enddo
            if (kk .lt. bsize) then
                y(i+kk) = y(i+kk) + a(acol) * x(col+kk)
            endif
           endif
	   i = k+1
        endif
      enddo
C
      return
      end
C
      integer function mvextra( ia, ja, a, n )
      implicit none
      integer n, ia(*), ja(*)
      double precision a(*)
      mvextra = n
      return
      end
C
      subroutine mvname( str )
      implicit none
      character*(*) str
      str = 'Fortran AIJ (unrolled) with Diagonal Block Detection'
      return
      end
C
      subroutine mvsetup( ia, ja, ka, a, n )
      implicit none
      integer ia(*), ja(*), ka(*)
      double precision a(*)
      integer i, n, pattern_row, offset, jpattern, npattern, cnt
      integer jrow, ncol, j
C 
      cnt = 1
C      /* Fill ka array with next row that has a different ja structure */
      pattern_row = 1
      offset      = 1
      jpattern = ia(pattern_row)
      npattern = ia(pattern_row+1) - jpattern
      do i = 2,n
  	jrow = ia(i) 
  	ncol = ia(i+1) - jrow
  	if (ncol .ne. npattern) then
          ka(cnt)   = i-1
          cnt = cnt+1
          pattern_row = i
          offset	    = 0
          jpattern    = ia(pattern_row)
          npattern    = ia(pattern_row+1) - jpattern
  	else 
            do j=1,ncol
               if (ja(jrow+j) .ne. ja(jpattern+j) + offset + 1 ) then
                  goto 10
               endif
            enddo
 10         continue
  	    if (j  .ne. ncol) then
  		ka(cnt)	= i-1
                cnt = cnt + 1
  		pattern_row	= i
  		offset		= 0
  		jpattern	= ia(pattern_row)
  		npattern	= ia(pattern_row+1) - jpattern
  	    else 
  		offset = offset + 1
  	    endif
  	endif
      enddo
      ka(cnt) = n 
      cnt = cnt+1
      return
      end
C
      subroutine mvflags( str )
      character*(*) str
      str = 'FLAGS'
      return
      end
