/* 
 * This version looks for regular increment patterns in the ja array.
 * The ka array basically gives the diagonal blocks of the matrix.
 * The data structure is otherwise AIJ; in particular, data has NOT
 * been re-ordered to improve diagonal access.
 */
#include "mv.h"

void mv_( ia, ja, ka,  a, x, y, n_ptr )
int    *ia, *ja, *ka, *n_ptr;
double *a, *x, *y;
{
  double     sum;
  int        i, jrow,j,ncol,shift,n=*n_ptr;
  int        k, cnt, kk, col, acol, bsize;

  /* x    = x + shift; */ 
  shift = 1; /* shift for Fortran start by 1 indexing */
  x = x - shift;
  i = 0;
  cnt = 0;
  while (i < n) {
    jrow  = ia[i];
    ncol  = ia[i+1] - jrow;
    jrow  = jrow - shift;
    k	  = ka[cnt++];
    bsize = k - i;
    if (bsize == 1) {
	/* If a single row, drop to regular code */
	sum  = 0.0;
/* Ignore vector dependencies */
#pragma ivdep
/* #pragma unroll (2) */
	for ( j=0; j<ncol; j++) {
	    sum += a[jrow]*x[ja[jrow]]; jrow++;
	}
	y[i] = sum;
	i++;
    }
    else {
	/* Switch to diagonal code */
	CHECKSUB(i+bsize,n);
	for (kk=0; kk<bsize; kk++) 
	    y[i+kk] = 0.0;
	for (j=0; j<ncol;j++) {
	    acol = jrow+j;
	    col  = ja[acol];
	    CHECKSUB(i+bsize,n);
/* #pragma prefetch_ref=a[acol], kind=rd, stride=number of iterations between */
	    for (kk=0; kk<bsize; kk++) {
		y[i+kk] += a[acol] * x[col+kk];
		acol    += ncol;
	    }
	}
	i = k;
    }
  }
}

int mvextra_( ia, ja, a, n_ptr )
int *ia, *ja, *n_ptr;
double *a;
{
    return *n_ptr + 1;
}

void mvsetup_( ia, ja, ka, a, n_ptr )
int *ia, *ja, *ka, *n_ptr;
double *a;
{
    int i, n, pattern_row, offset, jpattern, npattern, cnt;
    int jrow, ncol, j;

    n   = *n_ptr;
    cnt = 0;
    /* Fill ka array with next row that has a different ja structure */
    pattern_row = 0;
    offset      = 0;
    jpattern = ia[pattern_row];
    npattern = ia[pattern_row+1] - jpattern;
    for (i=1; i<n; i++) {
	jrow = ia[i];
	ncol = ia[i+1] - jrow;
	if (ncol != npattern) {
#ifdef DEBUG
	    printf( "failed number in col\n" );
#endif
	    CHECKSUB(cnt,n+1);
	    ka[cnt++]       = i;
	    pattern_row	    = i;
	    offset	    = 0;
	    jpattern	    = ia[pattern_row];
	    npattern	    = ia[pattern_row+1] - jpattern;
	}
	else {
	    /* Check each entry in ja */
	    for (j=0; j<ncol; j++) {
		if (ja[jrow+j] != ja[jpattern+j] + offset + 1) {
#ifdef DEBUG
		    printf( "failed column %d not %d\n",
			    ja[jrow+j], ja[jpattern+j] + offset + 1);
#endif
		    break;
		}
	    }
	    if (j != ncol) {
		CHECKSUB(cnt,n+1);
		ka[cnt++]	= i;
		pattern_row	= i;
		offset		= 0;
		jpattern	= ia[pattern_row];
		npattern	= ia[pattern_row+1] - jpattern;
	    }
	    else {
		offset++;
	    }
	}
    }
    CHECKSUB(cnt,n+1);
    ka[cnt++] = n;
#ifdef DEBUG
    for (i=0; i<cnt; i++) 
	printf( "ka[%d] = \t%d\n", i, ka[i] );
#endif
}

void mvname_( str, ln )
char *str;
int  ln;
{
    strncpy( str, "C AIJ with Diagonal Block Detection", ln - 1 );
}

#ifndef FLAGS
#define FLAGS "Unknown"
#endif
void mvflags_(str, ln )
char *str;
int  ln;
{
    strncpy( str, FLAGS, ln - 1 );
}
